from abc import ABC, abstractmethod
from multiprocessing import Queue
from queue import Empty
from threading import Event, Lock
from typing import Generator

import pika
from pika.adapters.blocking_connection import BlockingChannel, BlockingConnection
from pika.exceptions import AMQPError, ChannelError

from abstra_internals.environment import (
    EXECUTION_QUEUE_CONCURRENCY,
    RABBITMQ_EXECUTION_QUEUE,
    RABBITMQ_QUEUE_RECONNECT_ATTEMPTS,
    RABBITMQ_QUEUE_RECONNECT_BACKOFF_TIME,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.producer import PreExecution, QueueMessage
from abstra_internals.utils import deserialize


class Consumer(ABC):
    @abstractmethod
    def done_callback(self, msg: QueueMessage):
        raise NotImplementedError

    @abstractmethod
    def iter(self) -> Generator[QueueMessage, None, None]:
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError


class RabbitConsumer(Consumer):
    channel: BlockingChannel
    connection: BlockingConnection
    conn_uri: str
    is_connected: bool = False
    __connection_lock: Lock = Lock()

    def _create_params(self) -> pika.URLParameters:
        params = pika.URLParameters(self.conn_uri)
        params.retry_delay = RABBITMQ_QUEUE_RECONNECT_BACKOFF_TIME
        params.connection_attempts = RABBITMQ_QUEUE_RECONNECT_ATTEMPTS
        return params

    def __init__(self, conn_uri: str):
        self.conn_uri = conn_uri

        self.stop_evt = Event()
        self.concurrency = EXECUTION_QUEUE_CONCURRENCY
        self.queue = RABBITMQ_EXECUTION_QUEUE
        self._connect()

    def _connect(self):
        """Establish connection with retry logic"""
        with self.__connection_lock:
            while not self.stop_evt.is_set() and not self.is_connected:
                try:
                    params = self._create_params()
                    self.connection = pika.BlockingConnection(params)
                    self.channel = self.connection.channel()
                    self.channel.basic_qos(prefetch_count=self.concurrency)
                    self.channel.queue_declare(queue=self.queue, durable=True)
                    self.is_connected = True
                    AbstraLogger.info("Connected to RabbitMQ")
                    return
                except Exception as e:
                    AbstraLogger.error(f"Connection failed: {e}.")

    def _callback_threadsafe(self, msg: QueueMessage):
        with self.__connection_lock:
            if self.channel.is_open:
                AbstraLogger.debug(f"[CONSUMER] ack {msg.delivery_tag}")
                self.channel.basic_ack(msg.delivery_tag)
            else:
                AbstraLogger.capture_exception(
                    Exception(f"[CONSUMER] cannot ack {msg.delivery_tag}")
                )

    def done_callback(self, msg: QueueMessage):
        self.connection.add_callback_threadsafe(lambda: self._callback_threadsafe(msg))

    def iter(self) -> Generator[QueueMessage, None, None]:
        AbstraLogger.warning("[CONSUMER] starting")
        while True:
            if self.stop_evt.is_set():
                break

            try:
                for method, properties, body in self.channel.consume(
                    queue=self.queue, inactivity_timeout=10, auto_ack=False
                ):
                    if self.stop_evt.is_set():
                        break

                    if not method:
                        continue

                    yield QueueMessage(
                        preexecution=PreExecution(**deserialize(body.decode("utf-8"))),
                        delivery_tag=method.delivery_tag,
                    )

            except (ChannelError, AMQPError) as e:
                self.is_connected = False
                AbstraLogger.error(f"Unexpected error: {e}. Reconnecting...")
                self._connect()

    def stop(self):
        AbstraLogger.warning("[CONSUMER] stoping")
        self.stop_evt.set()
        try:
            with self.__connection_lock:
                if self.connection.is_open:
                    AbstraLogger.warning("[CONSUMER] channel canceling")
                    self.channel.cancel()
                    AbstraLogger.warning("[CONSUMER] channel closing")
                    self.channel.close()
                    AbstraLogger.warning("[CONSUMER] connection closing")
                    self.connection.close()
        except Exception as e:
            AbstraLogger.error(f"Shutdown error: {e}")

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.process_data_events(time_limit=10)
        self.stop()
        return False

    def __enter__(self) -> "RabbitConsumer":
        return self


class EditorConsumer(Consumer):
    def __init__(self, queue: Queue) -> None:
        self.stop_evt = Event()
        self.queue = queue

    def done_callback(self, msg: QueueMessage):
        pass

    def stop(self):
        self.stop_evt.set()

    def iter(self) -> Generator[QueueMessage, None, None]:
        while True:
            if self.stop_evt.is_set():
                break

            try:
                if msg := self.queue.get(timeout=1):
                    yield msg
            except Empty:
                continue



import re
from typing import Callable, List, Optional

from abstra_internals.entities.forms.widgets.response_types import PhoneResponse
from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class PhoneInput(InputWidget):
    """Phone number input widget with country code selection.

    Attributes:
        value (PhoneResponse): The phone number entered by the user.
    """

    type = "phone-input"
    value: PhoneResponse

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        placeholder: Optional[str] = None,
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = False,
        disabled: bool = False,
        errors: Optional[str] = None,
        invalid_message: Optional[str] = None,
    ):
        """Initialize a PhoneInput widget.

        Args:
            label (str): Text label displayed above the input.
            key (Optional[str]): Identifier for the widget, defaults to label if not provided.
            placeholder (Optional[str]): Placeholder text displayed when the input is empty.
            required (bool): Whether the input must be filled before proceeding.
            hint (Optional[str]): Help text displayed below the input.
            full_width (bool): Whether the input should take up the full width of its container.
            disabled (bool): Whether the input is non-interactive.
            errors (Optional[str]): Pre-defined validation error messages to display.
            invalid_message (Optional[str]): Custom error message for invalid phone numbers.
        """
        self.label = label
        self.key = key or label
        self.placeholder = placeholder
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.disabled = disabled
        self.value = self.make_empty()
        self.errors = errors
        self.invalid_message = invalid_message

    def make_empty(self):
        return PhoneResponse(raw="", masked="", country_code="", national_number="")

    def _render(self):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "value": self._serialize_value(),
            "placeholder": self.placeholder,
            "required": self.required,
            "hint": self.hint,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "errors": self.errors,
            "invalidMessage": self.invalid_message,
        }

    @property
    def _validators(self) -> List[Callable[[], List[str]]]:
        return [self._validate_phone]

    def _validate_phone(self):
        if self.required and (
            self.value.country_code == "" or self.value.national_number == ""
        ):
            return ["i18n_error_required_field"]

        return []

    def _serialize_value(self):
        if isinstance(self.value, dict):
            return {
                "countryCode": re.sub(
                    "\\D", "", str(self.value.get("country_code", ""))
                ),
                "nationalNumber": re.sub(
                    "\\D", "", str(self.value.get("national_number", ""))
                ),
            }
        if isinstance(self.value, PhoneResponse):
            return {
                "countryCode": re.sub("\\D", "", str(self.value.country_code)),
                "nationalNumber": re.sub("\\D", "", str(self.value.national_number)),
            }
        return {"countryCode": "", "nationalNumber": ""}

    def _parse_value(self, value) -> PhoneResponse:
        if value is None:
            return self.make_empty()
        masked = f"+{value.get('countryCode')} {value.get('nationalNumber')}"
        raw = re.sub("\\D", "", masked)
        return PhoneResponse(
            raw=raw,
            masked=masked,
            country_code=f"+{value.get('countryCode')}",
            national_number=value.get("nationalNumber"),
        )



from typing import Dict, List, Optional, TypeVar, Union

from abstra_internals.controllers.sdk.sdk_ai import Format, Prompt
from abstra_internals.controllers.sdk.sdk_context import SDKContextStore

T = TypeVar("T")


def to_list(value: Union[T, List[T], None]) -> List[T]:
    if value is None:
        return []
    if isinstance(value, list):
        return value
    return [value]


def normalize_format(format: Dict[str, object]) -> Dict[str, object]:
    for key, value in format.items():
        if isinstance(value, str):
            format[key] = {"type": value}
        elif isinstance(value, list):
            format[key] = {"enum": value}
        elif value is bool:
            format[key] = {"type": "boolean"}
        elif value is int:
            format[key] = {"type": "integer"}
        elif value is float:
            format[key] = {"type": "number"}
        elif value is str:
            format[key] = {"type": "string"}
        else:
            format[key] = value
    return format


def prompt(
    prompt: Union["Prompt", List["Prompt"]],
    instructions: Union[str, List[str]] = [],
    format: Optional["Format"] = None,
    temperature: float = 1.0,
):
    """
    Send a prompt to the AI and get a response.

    Args:
        prompt (Union[Prompt, List[Prompt]]): The prompt(s) to send to the AI model.
        instructions (Union[str, List[str]]): Additional instructions for the AI. Defaults to [].
        format (Optional[Format]): The expected format for the AI response. Defaults to None.
        temperature (float): Controls randomness in the AI response, from 0.0 to 2.0. Defaults to 1.0.

    Returns:
        The AI response formatted according to the specified format if provided.

    Raises:
        ValueError: If temperature is not between 0.0 and 2.0.
    """
    if temperature < 0.0 or temperature > 2.0:
        raise ValueError("Temperature must be between 0.0 and 2.0")

    instructions_list = to_list(instructions)
    normalized_format = normalize_format(format) if format else None
    prompt_list = to_list(prompt)

    return SDKContextStore.get_by_thread().ai_sdk.prompt(
        prompt_list, instructions_list, normalized_format, temperature
    )



import json
from typing import Any

from abstra_internals.entities.forms.widgets.widget_base import OutputWidget


class PlotlyOutput(OutputWidget):
    """Plotly figure output widget for displaying interactive charts.

    Attributes:
        fig (Any): The Plotly figure to display.
    """

    type = "plotly-output"

    def __init__(
        self,
        fig: Any = None,
        *,
        label: str = "",
        key: str = "",
        full_width: bool = False,
    ):
        """Initialize a PlotlyOutput widget.

        Args:
            fig (Any): The Plotly figure to display.
            label (str): Text label displayed above the chart.
            key (str): Identifier for the widget.
            full_width (bool): Whether the chart should take up the full width of its container.
        """
        self.fig = fig
        self.label = label
        self.key = key
        self.full_width = full_width

    def serialize_figure(self):
        if self.fig is None:
            import plotly.express as px

            df = px.data.tips()
            fig = px.density_heatmap(df, x="total_bill", y="tip")
            fig_json = fig.to_json()
            if not isinstance(fig_json, str):
                raise Exception("fig.to_json() did not return a string")
            return json.loads(fig_json)
        return json.loads(self.fig.to_json())

    def _render(self):
        return {
            "type": self.type,
            "figure": self.serialize_figure(),
            "fullWidth": self.full_width,
            "label": self.label,
        }



import threading
import webbrowser

from abstra_internals.environment import HOST
from abstra_internals.settings import Settings


def browser_open_editor():
    threading.Timer(
        1, lambda: webbrowser.open(f"http://{HOST}:{Settings.server_port}/_editor")
    ).start()
